package contextG

/**
 * Environment for defining behavior and states within a layer.
 * The main propose is to provide virtual (layered) properties 
 * which can be accessed inside of the definition closure.
 */
class DefinitionEnvironment {
    // The layer and the class context of this environment  
    def layer, declaringClass
    
    /**
     * Executes the closure in the context of this environment. 
     * The context is primarily the layer but also the declaring class.
     */
    def applyDefinition(closure) {
        def oldDelegate = closure.delegate
        def oldResolveStrategy = closure.getResolveStrategy()
        // change the delegate of this closure to this environment 
        // -> properties and methods are lookup in the environment
        closure.delegate = this
        closure.resolveStrategy = Closure.DELEGATE_FIRST
        // execute it
        closure.call()
        // restore previous state of closure
        closure.delegate = oldDelegate
        closure.resolveStrategy = oldResolveStrategy
        
        // ensures that the metaclass of the given class is LayeredObjectMetaClass
        if(!(declaringClass.metaClass instanceof LayeredObjectMetaClass))
            GroovySystem.getMetaClassRegistry().setMetaClass(declaringClass, 
                new LayeredObjectMetaClass(declaringClass.metaClass))
    }

    /**
     * If properties are modified within the definition, it will be saved in the layer as a virtual property
     */
    void setProperty(String name, Object value) {
        layer.setProperty(declaringClass, name, value)
    }
    
    /**
     * Returns any (virtual) property or if none exists an method stub for easy expansion (compare to ExpandoMetaClass).
     */
    def getProperty(String name) {
        def prop = layer.getProperty(declaringClass, name)
        if(prop != null) prop = new ExpandableMethodStub(name: name, environment: this)
        prop
    }
}

/**
 * An ExpandableMethodStub represents a method stub with a name to which different implementations can be streamed.
 * The streamed closures should have different parameter lists to ensure unique signatures. 
 */
class ExpandableMethodStub {
    def name, environment
    
    def leftShift(Closure closure) {
        // Adds the closure to the layer
        environment.@layer.setProperty(environment.@declaringClass, name, closure)
		this
	}
}