package contextG

import org.codehaus.groovy.runtime.MetaClassHelper
import org.codehaus.groovy.runtime.metaclass.ClosureMetaMethod
import org.codehaus.groovy.runtime.DefaultGroovyMethods

/**
 * Saves properties of a method call. Used for proceed().
 */
class LayeredMethodCall {
    def klass, layer, method, args
}

/**
 * Extends the default MetaClass with dynamic method and property lookup in the active layers.
 */
class LayeredObjectMetaClass extends DelegatingMetaClass {
    def LayeredObjectMetaClass(oldMetaClass) {
        super(oldMetaClass)
    }
    
    /**
     * Overrides default behavior for method lookup in active layers.
     */
    def invokeMethod(Object object, String methodName, Object[] arguments) {
        invokeLayeredMethod(object, methodName, arguments, false)
    }

    /**
     * Overrides default behavior for method lookup in active layers.
     */
    def invokeMethod(Class sender, Object receiver, String methodName, Object[] arguments, boolean isCallToSuper, boolean fromInsideClass) {
        invokeLayeredMethod(receiver, methodName, arguments, false)
    }

    /**
     * Overrides default behavior for method lookup in active layers.
     */
    def invokeStaticMethod(Object object, String methodName, Object[] arguments) {
        invokeLayeredMethod(object, methodName, arguments, true)
    }

    /**
     * Overrides default behavior for property lookup in active layers.
     */
    Object getProperty(Object object, String property) {
        // checks if the is a layered property and returns it
        def overriddenProperty = Layer.getOverridingProperty(theClass, property)
        if(!Layer.NO_SUCH_PROPERTY.is(overriddenProperty)) return overriddenProperty
        // default property lookup in unlayered class
        super.getProperty(object, property)
    }

    /**
     * Overrides default behavior for property lookup in active layers.
     */
    void setProperty(Object object, String property, Object value) { 
         // checks if the is a layered property and overrides it
        if(!Layer.setOverridingProperty(theClass, property, value)) 
            // default property lookup in unlayered class
            super.setProperty(object, property, value)
    }
    
    /**
     * True if it is a proceed call.
     */
    def isProceed(methodName, arguments) {
        (methodName == "proceed") && !arguments        
    }
        
    /**
     * Invokes the method in the context of currently activated layers.
     * A thread-local call stack is maintained in order to support "proceed".
     */
    def methodCallStack = [ initialValue: { new Stack() } ] as ThreadLocal  
    def invokeLayeredMethod(Object object, String methodName, Object[] arguments, isStatic) {
        def lastLayer = null
        def klass = object instanceof Class ? object : object.class

		// if the method call is "proceed()", load the properties of the previous method call
        def stack = methodCallStack.get()
        if(isProceed(methodName, arguments) && stack) {
            lastLayer = stack.peek().layer
            klass = stack.peek().klass
            methodName = stack.peek().method
            arguments = stack.peek().args
        }
        
        // look for an overriden method
        // if method call is "proceed()", lastLayer indicates which layer was used last
        // all layers preceding lastLayer and lastLayer itself are then ignored
        def layer = findOverridingLayer(klass, methodName, arguments, lastLayer)
        def method = layer?.findMethod(klass, methodName, arguments, true)
        		
        // now call the closure in the context of the object, also maintain a method call stack
        if(method instanceof ClosureMetaMethod) {
            method.closure.delegate = object
            method.closure.resolveStrategy = Closure.DELEGATE_FIRST
            stack.push(new LayeredMethodCall(layer: layer, klass: klass, method: methodName, args: arguments))
            def result = method.invoke(object, arguments)
            stack.pop()
            return result
        }
        
        // if no overridden method was found, call the default implementations
        if(isStatic) return delegate.invokeStaticMethod(object, methodName, arguments)
        delegate.invokeMethod(object, methodName, arguments)
    }

    def findOverridingLayer(klass, methodName, arguments, lastLayer) {
    	findOverridingLayer(klass, methodName, MetaClassHelper.convertToTypeArray(arguments), lastLayer)
    }
    
    def findOverridingLayer(klass, methodName, Class[] argumentClasses, lastLayer) {
    	def originalMethod = delegate.pickMethod(methodName, argumentClasses)
    	def baseClass = originalMethod?.declaringClass?.cachedClass
    	while(!DefaultGroovyMethods.is(klass, null) && !DefaultGroovyMethods.is(klass, baseClass)) {
    		def layer = Layer.getOverridingLayer(klass, methodName, argumentClasses, lastLayer)
    		if(layer != null) return layer
    		klass = klass.superclass
    	}
    	if(klass != null)
    		return Layer.getOverridingLayer(klass, methodName, argumentClasses, lastLayer)
    	null
    }
    
    /**
     * PickMethod is mainly used inside of closures and is send to the metaClass of the delegate.
     * It provides a somewhat faster method invocation mechanism as unnecessary exception creation 
     * can be avoided when a method is not found in one of the many context-defining meta classes.
     */
    MetaMethod pickMethod(String methodName, Class[] arguments) {
    	// Currently, this shortcut is not supported as it is not clear how "proceed()" should deal with it
    	// because there is no garantuee that the picked method is invoked.
    	// Whether or not it is invoked cannot be determined unless we create a special wrapper around the MetaMethod.
    	// However, this results in a certain overhead which may or may not be bigger than calling invokeMethod itself.
    	// For the sake of simplicity, for now pickMethod always returns null which causes current Groovy
    	// implementations to use invokeMethod.
    	// This ensures a consistent behavior even if some caching mechanisms depending on pickMethod become useless.
    	
    	null
        /*def layer
        def stack = methodCallStack.get()
		// if the method call is "proceed()", load the properties of the previous method call
        if(isProceed(methodName, arguments) && stack) {
            def klass = stack.peek().klass
            methodName = stack.peek().method
            arguments = MetaClassHelper.convertToTypeArray(stack.peek().args)
            def lastLayer = stack.peek().layer
            // look for the next layer which overrides the method
            layer = Layer.getOverridingLayer(klass, methodName, arguments, lastLayer)
        }
        // look for the first layer which overrides the method
        else layer = Layer.getOverridingLayer(theClass, methodName, arguments, null)
        
        // if a layer with overridden implementation was found, pick the method from it
        if(layer)
            return layer?.pickMethod(theClass, methodName, arguments)
            		
        // invoke default behavior if no layered method was found
        delegate.pickMethod(methodName, arguments)*/
    }
}
